/* DChub - a Direct Connect Hub for Linux
 * Copyright (C) 2001 Eric Prevoteau
 *
 * do_hub_reg.c: Copyright (C) Eric Prevoteau <www@ac2i.tzo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <glib.h>

#include "do_hub_reg.h"
#include "main.h"
#include "cnx_lst.h"
#include "gvar.h"
#include "key.h"
#include "network.h"

/*************************/
/* output: 0=ok, 1=error */
/*************************/
static int send_this(int sck, GString *out)
{
	int ln;
#ifndef MSG_NOSIGNAL
	void (*old_handler)(int);
#endif

#ifdef MSG_NOSIGNAL
	ln=send(sck, out->str,out->len,MSG_NOSIGNAL);
#else
	old_handler = signal (SIGPIPE, SIG_IGN);
	ln=send(sck, out->str,out->len,0);
	signal (SIGPIPE, old_handler);
#endif
	if(ln==-1)
	{
		return 1;
	}
	if(ln==0)
	{
		return 1;
	}
	return 0;
}


static void perform_registration(char *main_server_addr, char *this_serv)
{
	static int sck=-1;
	char buf[512];
	int ln;
	char *t;
	struct sockaddr_in lcl;
	int len_lcl=sizeof(lcl);
	unsigned int lport;
	unsigned char magic;

	GString *in,*out;
	
#ifdef DEBUG
	printf("hub registration in progress\n");
#endif

	if(sck!=-1)
	{
		shutdown(sck,2);
		close(sck);
		sck=-1;
	}

	sck=create_and_open_sock_on(main_server_addr,2501,0,gl_bind_addr);
	if(sck<0)
	{
		perror("create - registration failed");
		return;
	}

#ifdef DEBUG
	printf("hub registration (1)\n");
#endif

	if(getsockname(sck,(void*)&lcl,&len_lcl)!=0)
	{
		perror("getsockname - registration failed");
		close(sck);
		return;
	}

	lport=ntohs(lcl.sin_port);

	magic=lport+(lport>>8);		/* add the 2 bytes of the local port and create a 1 byte magic value */
	

	{
#ifndef MSG_NOSIGNAL
		void (*old_handler)(int);

		old_handler = signal (SIGPIPE, SIG_IGN);
		ln=recv(sck, buf,sizeof(buf)-1,0);
		signal (SIGPIPE, old_handler);
#else
		ln=recv(sck, buf,sizeof(buf)-1,MSG_NOSIGNAL);
#endif
	}
#ifdef DEBUG
	printf("hub registration (2)\n");
#endif

	if(ln==-1)
	{
		perror("recv - registration failed");
		close(sck);
		return;
	}
	if(ln==0)
	{
		printf("recv no data - registration failed");
		close(sck);
		return;
	}

	buf[ln]='\0';

	in=g_string_new(buf);

	if(strncmp(buf,"$Lock ",6))
	{
		printf("no $Lock: %s",in->str);
		g_string_free(in,TRUE);
		close(sck);
		return;
	}

	in=g_string_erase(in,0,6);

	t=strstr(in->str," Pk=");
	if(t==NULL)
	{
		printf("no Pk= %s",in->str);
		g_string_free(in,TRUE);
		close(sck);
		return;
	}

	in=g_string_truncate(in,t-in->str);

#ifdef DEBUG
	printf("hub registration (3)\n");
#endif

	out=compute_hub_reg_access_key(in,magic);

	out=g_string_prepend(out,"$Key ");
	out=g_string_append_c(out,'|');
	out=g_string_append(out,this_serv);
	send_this(sck,out);

	g_string_free(in,TRUE);
	g_string_free(out,TRUE);

#ifdef DEBUG
	printf("hub registration (4)\n");
#endif

	printf("hub registration done\n");
#if 0
	sleep(DELAY_BETWEEN_REGISTRATION-60);
#endif
	return;
}

G_LOCK_DEFINE_STATIC(hub_reg_in_progress);

#include "color_debug.h"
/***************************/
/* hub registration thread */
/***************************/
static void *do_hub_reg(void *ptr)
{
	char *hubname;
	char *hubaddr;
	char *hubdesc;
	char *reg_serv;
	unsigned long long shared_size;
	unsigned int nb_users;

	/* printf("starting hub registration thread\n"); */
	if(G_TRYLOCK(hub_reg_in_progress)==TRUE)
	{
		G_LOCK(conf_file);
		hubname=e_db_str_get(conf_file,"HUBNAME");
		hubaddr=e_db_str_get(conf_file,"HUBADDR");
		hubdesc=e_db_str_get(conf_file,"HUBDESC");
		reg_serv=e_db_str_get(conf_file,"REG_SERV");
		G_UNLOCK(conf_file);
	
		if(reg_serv!=NULL)
		{
			GString *fhd;
#ifdef __CYGWIN__
			char		long_int[48];
#endif
			HL_LOCK_READ(cnx_struct_lock);
			get_hub_user_stat(&shared_size,&nb_users);
			HL_UNLOCK_READ(cnx_struct_lock);

			fhd=g_string_new("");
			g_string_sprintfa(fhd,"%s|%s",
								hubname!=NULL?hubname:"",
								hubaddr!=NULL?hubaddr:"");
			if (reg_hub_port == 0)
			{
				if(hub_port!=411)
					g_string_sprintfa(fhd,":%hu",hub_port);
			}
			else
			{
				if(reg_hub_port!=411)
					g_string_sprintfa(fhd,":%hu",reg_hub_port);
			}
#ifdef __CYGWIN__
			sprintf(long_int, "%Lu", shared_size);
			g_string_sprintfa(fhd,"|%s|%u|%s|",
								hubdesc!=NULL?hubdesc:"",
								nb_users,
								long_int);
#else
			g_string_sprintfa(fhd,"|%s|%u|%Lu|",
								hubdesc!=NULL?hubdesc:"",
								nb_users,
								shared_size);
#endif
			perform_registration(reg_serv,fhd->str);
			printf("registrer : "COLOGREEN1"%s"COLONONE"\n", fhd->str);
			g_string_free(fhd,TRUE);
		}
	
		if(hubname!=NULL)
			free(hubname);
		if(hubaddr!=NULL)
			free(hubaddr);
		if(hubdesc!=NULL)
			free(hubdesc);
		if(reg_serv!=NULL)
			free(reg_serv);
		G_UNLOCK(hub_reg_in_progress);
	}
	else
	{
		printf("earlier hub registration still in progress.\n");
	}
	pthread_exit(NULL);
	return(NULL);
}

/***********************************************/
/* check hub registration on DC hublist server */
/***********************************************/
void check_hub_registration(void)
{
	static time_t last_hub_registration=0;
	time_t cur_time;
	int do_hub_registration=0;

	G_LOCK(conf_file);
	e_db_int_get(conf_file,"REG_HUB",&do_hub_registration);
	G_UNLOCK(conf_file);

	if(do_hub_registration==0)
		return;

	cur_time=time(NULL);
	if((cur_time-last_hub_registration)>DELAY_BETWEEN_REGISTRATION)
	{
		static pthread_t thread_id;			/* these 2 variables must exist as long as the thread exists */
		static pthread_attr_t thread_attr;

		pthread_attr_init (&thread_attr);
		pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
		if(pthread_create(&thread_id,&thread_attr,(void*)do_hub_reg,NULL)!=0)
		{
			last_hub_registration+=10;		 /* fail -> retry 10 seconds later */
		}
		else
			last_hub_registration=cur_time;
	}
}


